@@ -1,4 +1,4 @@
-// Copyright (c) 2016-2020 The http-serve developers
+// Copyright (c) 2016-2021 The http-serve developers
 //
 // Licensed under the Apache License, Version 2.0 <LICENSE-APACHE.txt or
 // http://www.apache.org/licenses/LICENSE-2.0> or the MIT license
@@ -6,9 +6,9 @@
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
 
-//! Helpers for serving HTTP GET and HEAD responses with [hyper](https://crates.io/crates/hyper)
-//! 0.14.x and [tokio](https://crates.io/crates/tokio). A future version is likely to switch to
-//! the interface of the [http](http://crates.io/crates/http) crate.
+//! Helpers for serving HTTP GET and HEAD responses asynchronously with the
+//! [http](http://crates.io/crates/http) crate and [tokio](https://crates.io/crates/tokio).
+//! Works well with [hyper](https://crates.io/crates/hyper) 0.14.x.
 //!
 //! This crate supplies two ways to respond to HTTP GET and HEAD requests:
 //!
@@ -16,32 +16,61 @@
 //!     byte-rangeable HTTP entities. `Entity` must be able to produce exactly the same data on
 //!     every call, know its size in advance, and be able to produce portions of the data on demand.
 //! *   the `streaming_body` function can be used to add a body to an otherwise-complete response.
-//!     If a body is needed, it returns a `BodyWriter` (which implements `std::io::Writer`). The
-//!     caller should produce the complete body or call `BodyWriter::abort`, causing the HTTP
-//!     stream to terminate abruptly.
+//!     If a body is needed (on `GET` rather than `HEAD` requests), it returns a `BodyWriter`
+//!     (which implements `std::io::Writer`). The caller should produce the complete body or call
+//!     `BodyWriter::abort`, causing the HTTP stream to terminate abruptly.
+
+//! It supplies a static file `Entity` implementation and a (currently Unix-only)
+//! helper for serving a full directory tree from the local filesystem, including
+//! automatically looking for `.gz`-suffixed files when the client advertises
+//! `Accept-Encoding: gzip`.
 //!
 //! # Why two ways?
 //!
-//! They have pros and cons. This chart shows some of them:
+//! They have pros and cons. This table shows some of them:
 //!
 //! <table>
 //!   <tr><th><th><code>serve</code><th><code>streaming_body</code></tr>
-//!   <tr><td>automatic byte range serving<td>yes<td>no (always sends full body)</tr>
-//!   <tr><td>backpressure<td>yes<td>no</tr>
-//!   <tr><td>conditional GET<td>yes<td>unimplemented (always sends body)</tr>
+//!   <tr><td>automatic byte range serving<td>yes<td>no [<a href="#range">1</a>]</tr>
+//!   <tr><td>backpressure<td>yes<td>no [<a href="#backpressure">2</a>]</tr>
+//!   <tr><td>conditional GET<td>yes<td>no [<a href="#conditional_get">3</a>]</tr>
 //!   <tr><td>sends first byte before length known<td>no<td>yes</tr>
-//!   <tr><td>automatic gzip content encoding<td>no<td>yes</tr>
+//!   <tr><td>automatic gzip content encoding<td>no [<a href="#gzip">4</a>]<td>yes</tr>
 //! </table>
 //!
+//! <a name="range">\[1\]</a>: `streaming_body` always sends the full body. Byte range serving
+//! wouldn't make much sense with its interface. The application will generate all the bytes
+//! every time anyway, and `http-serve`'s buffering logic would have to be complex
+//! to handle multiple ranges well.
+//!
+//! <a name="backpressure">\[2\]</a>: `streaming_body` is often appended to while holding
+//! a lock or open database transaction, where backpressure is undesired. It'd be
+//! possible to add support for "wait points" where the caller explicitly wants backpressure. This
+//! would make it more suitable for large streams, even infinite streams like
+//! [Server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events).
+//!
+//! <a name="conditional_get">\[3\]</a>: `streaming_body` doesn't yet support
+//! generating etags or honoring conditional GET requests. PRs welcome!
+//!
+//! <a name="gzip">\[4\]</a>: `serve` doesn't automatically apply `Content-Encoding:
+//! gzip` because the content encoding is a property of the entity you supply. The
+//! entity's etag, length, and byte range boundaries must match the encoding. You
+//! can use the `http_serve::should_gzip` helper to decide between supplying a plain
+//! or gzipped entity. `serve` could automatically apply the related
+//! `Transfer-Encoding: gzip` where the browser requests it via `TE: gzip`, but
+//! common browsers have
+//! [chosen](https://bugs.chromium.org/p/chromium/issues/detail?id=94730) to avoid
+//! requesting or handling `Transfer-Encoding`.
+//!
 //! Use `serve` when:
 //!
 //! *   metadata (length, etag, etc) and byte ranges can be regenerated cheaply and consistently
-//!     via a lazy `Entity`.
+//!     via a lazy `Entity`, or
 //! *   data can be fully buffered in memory or on disk and reused many times. You may want to
 //!     create a pair of buffers for gzipped (for user-agents which specify `Accept-Encoding:
 //!     gzip`) vs raw.
 //!
-//! Consider `streaming_body` if data would be fully buffered each time a response is sent.
+//! Use `streaming_body` when regenerating the entire body each time a response is sent.
 //!
 //! Once you return a `hyper::server::Response` to hyper, your only way to signal error to the
 //! client is to abruptly close the HTTP connection while sending the body. If you want the ability
@@ -53,15 +82,31 @@
 //! similar). `streaming_body` doesn't need to keep its own copy for potential future use; it may
 //! be cheaper because it can simply hand ownership of the existing `Vec<u8>`s to hyper.
 //!
-//! # Why the weird type bounds? Why not use `hyper::Body` and `hyper::Chunk` for everything?
+//! # Why the weird type bounds? Why not use `hyper::Body` and `bytes::Bytes` for everything?
 //!
-//! These bounds are compatible with `hyper::Body` and `hyper::Chunk`, and most callers will use
-//! those types. There are times when it's desirable to have more flexible ownership provided by a
-//! type such as `reffers::ARefs<'static, [u8]>`. One is `mmap`-based file serving: a
-//! `hyper::Chunk` would require copying the data in each chunk. An implementation with `ARefs`
+//! These bounds are compatible with `hyper::Body` and `bytes::Bytes`, and most callers will use
+//! those types. **Note:** if you see an error like the one below, ensure you are using hyper's
+//! `stream` feature:
+//!
+//! ```text
+//! error[E0277]: the trait bound `Body: From<Box<(dyn futures::Stream<Item = Result<_, _>> +
+//! std::marker::Send + 'static)>>` is not satisfied
+//! ```
+//!
+//! `Cargo.toml` should look similar to the following:
+//!
+//! ```toml
+//! hyper = { version = "0.14.7", features = ["stream"] }
+//! ```
+//!
+//! There are times when it's desirable to have more flexible ownership provided by a
+//! type such as `reffers::ARefs<'static, [u8]>`. One is `mmap`-based file serving:
+//! `bytes::Bytes` would require copying the data in each chunk. An implementation with `ARefs`
 //! could instead `mmap` and `mlock` the data on another thread and provide chunks which `munmap`
 //! when dropped. In these cases, the caller can supply an alternate implementation of the
-//! `http_body::Body` trait which uses a different `Data` type than `hyper::Chunk`.
+//! `http_body::Body` trait which uses a different `Data` type than `bytes::Bytes`.
+
+#![cfg_attr(docsrs, feature(doc_cfg))]
 
 use bytes::Buf;
 use futures_core::Stream;
@@ -88,6 +133,7 @@ macro_rules! unsafe_fmt_ascii_val {
 mod chunker;
 
 #[cfg(feature = "dir")]
+#[cfg_attr(docsrs, doc(cfg(feature = "dir")))]
 pub mod dir;
 
 mod etag;
@@ -237,6 +283,7 @@ pub fn should_gzip(headers: &HeaderMap) -> bool {
     gzip_q > 0 && gzip_q >= identity_q
 }
 
+/// A builder returned by [streaming_body].
 pub struct StreamingBodyBuilder {
     chunk_size: usize,
     gzip_level: u32,
@@ -244,11 +291,50 @@ pub struct StreamingBodyBuilder {
     body_needed: bool,
 }
 
-/// Adds a streaming body to the given request if a body is needed.
+/// Creates a response and streaming body writer for the given request.
 ///
-/// Currently the body is added for non-HEAD requests. In the future, this may also follow
-/// conditional GET rules, omitting the body and stripping out entity headers from the response as
-/// desired.
+/// The streaming body writer is currently `Some(writer)` for `GET` requests and
+/// `None` for `HEAD` requests. In the future, `streaming_body` may also support
+/// conditional `GET` requests.
+///
+/// ```
+/// # use http::{Request, Response, header::{self, HeaderValue}};
+/// use std::io::Write as _;
+///
+/// fn respond(req: Request<hyper::Body>) -> std::io::Result<Response<hyper::Body>> {
+///     let (mut resp, stream) = http_serve::streaming_body(&req).build();
+///     if let Some(mut w) = stream {
+///         write!(&mut w, "hello world")?;
+///     }
+///     resp.headers_mut().insert(header::CONTENT_TYPE, HeaderValue::from_static("text/plain"));
+///     Ok(resp)
+/// }
+/// ```
+///
+/// The caller may also continue appending to `stream` after returning the response to `hyper`.
+/// The response will end when `stream` is dropped. The only disadvantage to writing to the stream
+/// after the fact is that there's no way to report mid-response errors other than abruptly closing
+/// the TCP connection ([BodyWriter::abort]).
+///
+/// ```
+/// # use http::{Request, Response, header::{self, HeaderValue}};
+/// use std::io::Write as _;
+///
+/// fn respond(req: Request<hyper::Body>) -> std::io::Result<Response<hyper::Body>> {
+///     let (mut resp, stream) = http_serve::streaming_body(&req).build();
+///     if let Some(mut w) = stream {
+///         tokio::spawn(async move {
+///             for i in 0..10 {
+///                 tokio::time::sleep(std::time::Duration::from_secs(1)).await;
+///                 write!(&mut w, "write {}\n", i)?;
+///             }
+///             Ok::<_, std::io::Error>(())
+///         });
+///     }
+///     resp.headers_mut().insert(header::CONTENT_TYPE, HeaderValue::from_static("text/plain"));
+///     Ok(resp)
+/// }
+/// ```
 pub fn streaming_body<T>(req: &http::Request<T>) -> StreamingBodyBuilder {
     StreamingBodyBuilder {
         chunk_size: 4096,
@@ -267,7 +353,7 @@ impl StreamingBodyBuilder {
         StreamingBodyBuilder { chunk_size, ..self }
     }
 
-    /// Sets the gzip compression level.
+    /// Sets the gzip compression level. Defaults to 6.
     ///
     /// `gzip_level` should be an integer between 0 and 9 (inclusive).
     /// 0 means no compression; 9 gives the best compression (but most CPU usage).
@@ -277,6 +363,7 @@ impl StreamingBodyBuilder {
         StreamingBodyBuilder { gzip_level, ..self }
     }
 
+    /// Returns the HTTP response and, if the request is a `GET`, a body writer.
     pub fn build<P, D, E>(self) -> (http::Response<P>, Option<BodyWriter<D, E>>)
     where
         D: From<Vec<u8>> + Send + Sync,
